from circuiteasy.complex_utils import polar_to_complex
import math 

def dc_power(v: float = None, i: float = None, r: float = None) -> float:
    
    given = [v is not None, i is not None, r is not None].count(True)
    if given != 2:
        raise ValueError("Provide exactly two of (v, i, r) to calculate power.")
    if v is not None and i is not None:
        return round(v * i, 2)
    if i is not None and r is not None:
        return round(i ** 2 * r, 2)
    if v is not None and r is not None:
        return round(v ** 2 / r, 2)

def ac_active_power(
    Vrms: float = None,
    Irms: float = None,
    R: float = None,
    pf: float = None,
    phase: float = None,
    S_mag: float = None) -> float:
    """
    Calculate AC active (real) power (P) using available values.

    Possible formulas:
        - P = Vrms * Irms * pf            (if Vrms, Irms, pf given)
        - P = Vrms * Irms * cos(phase)    (if Vrms, Irms, phase given)
        - P = S_mag * pf                  (if S_mag, pf given)
        - P = S_mag * cos(phase)          (if S_mag, phase given)
        - P = Vrms^2 / R                  (if Vrms, R given, for purely resistive load)
        - P = Irms^2 * R                  (if Irms, R given, for purely resistive load)
    
    Args:
        Vrms: RMS voltage (V)
        Irms: RMS current (A)
        R: Resistance (Ohm)
        pf: Power factor (cosϕ), 0...1
        phase: Phase angle in degrees (ϕ)
        S_mag: Apparent power magnitude (VA)
    
    Returns:
        Active power P (W), rounded to 2 decimals.
    """
    if phase is not None:
        phase_rad = math.radians(phase)
        if Vrms is not None and Irms is not None:
            return round(Vrms * Irms * math.cos(phase_rad), 2)
        if S_mag is not None:
            return round(S_mag * math.cos(phase_rad), 2)
    if Vrms is not None and Irms is not None and pf is not None:
        return round(Vrms * Irms * pf, 2)
    if S_mag is not None and pf is not None:
        return round(S_mag * pf, 2)
    if Vrms is not None and R is not None:
        return round(Vrms ** 2 / R, 2)
    if Irms is not None and R is not None:
        return round(Irms ** 2 * R, 2)
    raise ValueError("Provide (Vrms, Irms, pf), (Vrms, Irms, phase), (S_mag, pf), (S_mag, phase), (Vrms, R), or (Irms, R).")


import math

def ac_reactive_power(
    Vrms: float = None,
    Irms: float = None,
    S_mag: float = None,
    phase: float = None,
    Z: float = None,
    pf: float = None,
    Vrms_phasor: complex = None,
    Irms_conjphsor: complex = None) -> complex:
    """
    Calculate AC reactive power (Q) as a complex value using available values.

    Possible formulas:
        - Q = Vrms * Irms * sin(phase)          (if Vrms, Irms, phase given)
        - Q = S_mag * sin(phase)                (if S_mag, phase given)
        - Q = S_mag * sqrt(1 - pf^2)            (if S_mag, pf given)
        - Q = Vrms^2 / Z                        (if Vrms, Z given, for purely reactive)
        - Q = Irms^2 * Z                        (if Irms, Z given, for purely reactive)
        - Q = imag(Vrms_phasor * Irms_conjphsor)        (if Vrms_phasor, Irms_conjphsor given)
    
    Args:
        Vrms: RMS voltage (V)
        Irms: RMS current (A)
        S_mag: Apparent power magnitude (VA)
        phase: Phase angle in degrees (ϕ)
        Z: Reactance (Ohm)
        pf: Power factor (cosϕ), 0...1
        Vrms_phasor: RMS voltage as a phasor (complex, must be RMS)
        Irms_conjphsor: Conjugated current phasor (complex, must be RMS)
    
    Returns:
        Reactive power Q as a complex (0 + jQ), rounded to 2 decimals.
    """
    if phase is not None:
        phase_rad = math.radians(phase)
        if Vrms is not None and Irms is not None:
            Q = Vrms * Irms * math.sin(phase_rad)
            return complex(0, round(Q, 2))
        if S_mag is not None:
            Q = S_mag * math.sin(phase_rad)
            return complex(0, round(Q, 2))
    if S_mag is not None and pf is not None:
        if not (0 <= pf <= 1):
            raise ValueError("Power factor must be between 0 and 1.")
        Q = S_mag * math.sqrt(1 - pf**2)
        return complex(0, round(Q, 2))
    if Vrms is not None and Z is not None:
        Q = Vrms ** 2 / Z
        return complex(0, round(Q, 2))
    if Irms is not None and Z is not None:
        Q = Irms ** 2 * Z
        return complex(0, round(Q, 2))
    if Vrms_phasor is not None and Irms_conjphsor is not None:
        Q = (Vrms_phasor * Irms_conjphsor).imag
        return complex(0, round(Q, 2))
    raise ValueError("Provide (Vrms, Irms, phase), (S_mag, phase), (S_mag, pf), (Vrms, Z), (Irms, Z), or (Vrms_phasor, Irms_conjphsor) for calculation.")

def ac_apparent_power(P: float = None, Q: float = None, Vrms: float = None, Irms: float = None) -> complex:
    """
    Calculate complex apparent power S.
    Either supply (P, Q) or (Vrms, Irms).
    Returns S as complex number: S = P + jQ (units: VA)
    """
    if P is not None and Q is not None:
        return complex(round(P, 2), round(Q, 2))
    if Vrms is not None and Irms is not None:
        S = Vrms * Irms
        return complex(round(S, 2), 0)
    raise ValueError("Provide (P, Q) or (Vrms, Irms) for apparent power.")


def apparent_power_magnitude(Vrms: float = None, Irms: float = None, P: float = None, Q: float = None) -> float:
    """
    Calculate apparent power magnitude |S| = sqrt(P^2 + Q^2) or Vrms * Irms.
    """
    if Vrms is not None and Irms is not None:
        return round(Vrms * Irms, 2)
    if P is not None and Q is not None:
        return round(math.sqrt(P**2 + Q**2), 2)
    raise ValueError("Provide (Vrms, Irms) or (P, Q) for S magnitude.")


def ac_apparent_power_phasors(Vrms: complex = None, Irms_conj: complex = None, Z: complex = None) -> complex:
    """
    Calculate apparent (complex) power S using phasors.
    Args:
        Vrms      : Voltage phasor (RMS, complex)
        Irms_conj : Conjugated current phasor (RMS, complex)
        Z         : Complex impedance (optional, for alternate calc)
    Returns:
        Complex apparent power S (in VA)
    Raises:
        ValueError if required arguments are missing.
    """
    if Vrms is not None and Irms_conj is not None:
        return Vrms * Irms_conj
    if Vrms is not None and Z is not None:
        S_mag = abs(Vrms)**2 / abs(Z)
        V_angle = math.degrees(math.atan2(Vrms.imag, Vrms.real))
        Z_angle = math.degrees(math.atan2(Z.imag, Z.real))
        S_angle = 2 * V_angle - Z_angle
        S = polar_to_complex(S_mag, S_angle)
        return S
    raise ValueError("Provide (Vrms, Irms_conj) for S = V·I*, or (Vrms, Z) for |S| = |V|^2/|Z|.")


def print_complex_power(S: complex):
    """
    Print apparent power in P + jQ form and as magnitude.
    """
    P = S.real
    Q = S.imag
    mag = abs(S)
    print(f" S = ({P:.2f})+ ({Q:.2f}j VA) ")
    print(f"|S| = {mag:.2f} VA \n")

def db_power(Pout, Pin):
    """
    Returnerer effektforholdet i desibel (dB) basert på utgangs- og inngangseffekt.
    Brukes når du sammenligner to effekter (W).
    Formel: 10 * log10(Pout / Pin)
    """
    if Pin == 0:
        raise ValueError("Inngangseffekten kan ikke være null.")
    return 10 * math.log10(Pout / Pin)

def db_power(Pout, Pin):
    """
    Returnerer effektforholdet i desibel (dB) basert på utgangs- og inngangseffekt.
    Brukes når du sammenligner to effekter (W).
    Formel: 10 * log10(Pout / Pin)
    """
    if Pin == 0:
        raise ValueError("Inngangseffekten kan ikke være null.")
    return 10 * math.log10(Pout / Pin)

def db_current(Iout, Iin):
    """
    Returnerer strømforholdet i desibel (dB) basert på utgangs- og inngangsstrøm.
    Brukes når du sammenligner to strømmer (A).
    Formel: 20 * log10(Iout / Iin)
    """
    if Iin == 0:
        raise ValueError("Inngangsstrømmen kan ikke være null.")
    return 20 * math.log10(Iout / Iin)



from sympy import symbols, solve, sympify
import sympy as sp
import cmath
import math

def solve_equations(variables, equations, component_values):
    """
    Solve a system of equations symbolically (with SymPy) for circuit analysis.
    Args:
        variables        : List of variable names to solve for (strings).
        equations        : List of equations as strings (e.g., 'I1 + I2 - 2').
        component_values : Dictionary of {component_name: value} for substitution.
    Returns:
        dict with:
            'formatted': List of formatted solution strings (mag∠deg, real+j imag).
            'values'   : Dict {var: solution as complex}
    Example:
        solve_equations(['I1','I2'], ['I1+I2-2', '2*I1-I2-1'], {})
    """
    syms = symbols(variables)
    var_map = dict(zip(variables, syms))
    eqs = []
    for eq in equations:
        sym_eq = sympify(eq, locals=var_map)
        for comp, val in component_values.items():
            sym_eq = sym_eq.subs(var_map.get(comp, comp), val)
        eqs.append(sym_eq)
    sol = solve(eqs, syms, dict=True)
    if not sol:
        return None

    formatted = []
    values = {}
    for var in variables:
        val = sol[0][var_map[var]]
        try:
            val_c = complex(val.evalf())
        except Exception:
            val_c = val
        values[var] = val_c
        mag, angle_rad = cmath.polar(val_c)
        angle_deg = math.degrees(angle_rad)
        formatted.append(f"{var}: ({mag:.4f}) ∠ ({angle_deg:.4f}°)")
        formatted.append(f"{var}: ({val_c.real:.4f}) + ({val_c.imag:.4f})j\n")
    return {"formatted": formatted, "values": values}

def simplify(eqns, unknowns, values):
    """
    Take equations as strings, unknowns as strings, and values as a dict.
    Automates symbol creation, solves, substitutes, and simplifies.
    """
    # 1. Gather all unique variable names from equations and unknowns/values
    varnames = set()
    # from equations
    for eq in eqns:
        varnames.update([str(sym) for sym in sp.sympify(eq).free_symbols])
    # from unknowns & values
    varnames.update([str(u) for u in unknowns])
    varnames.update([str(k) for k in values.keys()])
    # 2. Create sympy symbols for all
    symbols = {name: sp.symbols(name) for name in varnames}
    # 3. Parse equations as sympy expressions, turning "=" into Eq
    parsed_eqns = []
    for eq in eqns:
        if "=" in eq:
            left, right = eq.split("=")
            parsed_eqns.append(sp.Eq(sp.sympify(left, locals=symbols), sp.sympify(right, locals=symbols)))
        else:
            parsed_eqns.append(sp.sympify(eq, locals=symbols))
    # 4. Build unknown symbols
    unknown_syms = [symbols[str(u)] for u in unknowns]
    # 5. Substitute values as sympy symbols
    sub_dict = {symbols[str(k)]: v for k, v in values.items()}
    # 6. Solve
    sol = sp.solve(parsed_eqns, unknown_syms, dict=True)
    if not sol:
        raise ValueError("No solution found!")
    sol = sol[0]
    # 7. Substitute and simplify
    results = []
    for var in unknowns:
        expr = sol[symbols[var]]
        expr_sub = expr.subs(sub_dict)
        expr_simp = sp.simplify(expr_sub)
        results.append(expr_simp)
    return tuple(results)



import sympy as sp

def laplace_transform_of(expr):
    """
    Compute the Laplace transform of an expression of t (symbolic or as string).
    Args:
        expr: Expression in time domain (as SymPy or string, using 't').
    Returns:
        Laplace transform as string (in 's' domain).
    Raises:
        ValueError or RuntimeError on bad input or failure.
    """
    t, s = sp.symbols('t s')

    if isinstance(expr, str):
        try:
            expr = sp.sympify(expr)
        except sp.SympifyError:
            raise ValueError("Input expression could not be parsed.")

    allowed_vars = {t}
    expr_vars = expr.free_symbols
    if not expr_vars.issubset(allowed_vars):
        raise ValueError(f"Unsupported variables present: {expr_vars - allowed_vars}")

    try:
        L = sp.laplace_transform(expr, t, s, noconds=True)
    except Exception as e:
        raise RuntimeError(f"Laplace transform calculation failed: {e}")

    return str(L)

def inverse_laplace_transform_of(expr):
    """
    Compute the inverse Laplace transform of an expression of s (symbolic or as string).
    Args:
        expr: Expression in s-domain (as SymPy or string, using 's').
    Returns:
        Inverse Laplace transform as string (in 't' domain).
    Raises:
        ValueError or RuntimeError on bad input or failure.
    """
    t, s = sp.symbols('t s')

    if isinstance(expr, str):
        try:
            expr = sp.sympify(expr)
        except sp.SympifyError:
            raise ValueError("Input expression could not be parsed.")

    allowed_vars = {s}
    expr_vars = expr.free_symbols
    if not expr_vars.issubset(allowed_vars):
        raise ValueError(f"Unsupported variables present: {expr_vars - allowed_vars}")

    try:
        inv_L = sp.inverse_laplace_transform(expr, s, t)
    except Exception as e:
        raise RuntimeError(f"Inverse Laplace transform calculation failed: {e}")

    return str(inv_L)



from setuptools import setup, find_packages

setup(
    name="circuiteasy",
    version="v1.0.5",
    description="ELE142 and general circuit analysis tools for fast, clear Python calculations.",
    author="Hareth Al-jomaa",
    author_email="harethaljomaa@outlook.com",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "matplotlib",
        "sympy"
    ],
    python_requires=">=3.7",
    include_package_data=True,
    license="MIT",
    url="https://github.com/haljoumaa/circuiteasy.git", 
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
